// jack.c
// LiVES (lives-exe)
// (c) G. Finch 2005 - 2008
// Released under the GPL 3 or later
// see file ../COPYING for licensing details

#include "main.h"
#include "audio.h"
#include "events.h"
#include "effects-weed.h"
#include "support.h"

inline void sample_silence_dS (float *dst, unsigned long nsamples) {
  memset(dst,0,nsamples*sizeof(float));
}


void sample_move_d8_d16(short *dst, unsigned char *src,
			unsigned long nsamples, float scale, int nDstChannels, int nSrcChannels) {
  register int nSrcCount, nDstCount;
  register float src_offset_f=0.f;
  register int src_offset_i=0;
  register int ccount;
  
  //TRACE("nsamples == %ld, nDstChannels == %d, nSrcChannels == %d\n", nsamples, nDstChannels, nSrcChannels);

  if(!nSrcChannels) return;

  if (scale<0.f) {
    src_offset_f=((float)(nsamples)*(-scale)-1.f);
    src_offset_i=(int)src_offset_f*nSrcChannels;
  }

  while(nsamples--) {
    nSrcCount = nSrcChannels;
    nDstCount = nDstChannels;
    ccount=0;

    /* loop until all of our destination channels are filled */
    while(nDstCount) {
      nSrcCount--;
      nDstCount--;

      *(dst++) = *(src+ccount+src_offset_i)<<8; /* copy the data over */
      ccount++;

      /* if we ran out of source channels but not destination channels */
      /* then start the src channels back where we were */
      if(!nSrcCount && nDstCount) {
	ccount=0;
	nSrcCount = nSrcChannels;
      }
    }
    
    /* advance the the position */
    src_offset_i=(int)(src_offset_f+=scale)*nSrcChannels;
  }
}



/* convert from any number of source channels to any number of destination channels */
void sample_move_d16_d16(short *dst, short *src,
			 unsigned long nsamples, float scale, int nDstChannels, int nSrcChannels, int swap_endian) {
  register int nSrcCount, nDstCount;
  register float src_offset_f=0.f;
  register int src_offset_i=0;
  register int ccount;

  if(!nSrcChannels) return;

  if (scale<0.f) {
    src_offset_f=((float)(nsamples)*(-scale)-1.f);
    src_offset_i=(int)src_offset_f*nSrcChannels;
  }

  while(nsamples--) {
    if ((nSrcCount = nSrcChannels)==(nDstCount = nDstChannels)&&!swap_endian) {
      // same number of channels
      w_memcpy(dst,src+src_offset_i,nSrcChannels*sizeof(short));
      dst+=nDstCount;
    } 
    else {
      ccount=0;

      /* loop until all of our destination channels are filled */
      while(nDstCount) {
	nSrcCount--;
	nDstCount--;
	
	if (!swap_endian) *(dst++) = *(src+ccount+src_offset_i); /* copy the data over */
	else *(dst++)=((*(src+ccount+src_offset_i)&0x00FF)<<8)+(*(src+ccount+src_offset_i)>>8);
	ccount++;
	
	/* if we ran out of source channels but not destination channels */
	/* then start the src channels back where we were */
	if(!nSrcCount && nDstCount) {
	  ccount=0;
	  nSrcCount = nSrcChannels;
	}
      }
    }
    /* advance the the position */
    src_offset_i=(int)(src_offset_f+=scale)*nSrcChannels;
  }
}

/* convert from 16 bit to floating point */
/* channels to a buffer that will hold a single channel stream */
/* src_skip is in terms of 16bit samples */
void sample_move_d16_float (float *dst, short *src, unsigned long nsamples, unsigned long src_skip, int is_unsigned, float vol) {
  register float svol;
  float val;

#ifdef ENABLE_OIL
  double x;
  double y=0.f;
#endif

  if (vol==0.) vol=0.0000001f;
  svol=SAMPLE_MAX_16BIT/vol;

#ifdef ENABLE_OIL
  x=1./svol;
#endif

  while (nsamples--) {
    if (!is_unsigned) {
#ifdef ENABLE_OIL
      oil_scaleconv_f32_s16(&val,src,1,&y,&x);
#else
      if ((val = (float)((*src) / svol))>1.0f) val=1.0f;
      else if (val<-1.0f) val=-1.0f;
#endif
    }
    else {
#ifdef ENABLE_OIL
      oil_scaleconv_f32_u16(&val,(unsigned short *)src,1,&y,&x);
#else
      if ((val=((float)((unsigned short)(*src)) / SAMPLE_MAX_16BIT - 1.0f)*vol)>1.0f) val=1.0f;
      else if (val<-1.0f) val=-1.0f;
#endif
    }
    *(dst++)=val;
    src += src_skip;
  }

}



long sample_move_float_int(void *holding_buff, float **float_buffer, int nsamps, float scale, int chans, int asamps, int usigned, int swap_endian, double vol) {
  // convert float samples back to int
  long frames_out=0l;
  short val;
  unsigned short valu=0;
  register int i;
  register int offs=0,coffs=0;
  register float coffs_f=0.f;
  short *hbuffs=(short *)holding_buff;
  unsigned short *hbuffu=(unsigned short *)holding_buff;
  unsigned char *hbuffc=(guchar *)holding_buff;

  while ((nsamps-coffs)>0) {
    frames_out++;
    for (i=0;i<chans;i++) {
      val=(short)(*(float_buffer[i]+coffs)*vol*SAMPLE_MAX_16BIT);
      if (usigned) {
	valu=val+SAMPLE_MAX_16BITI;
      }
      if (asamps==16) {
	if (!swap_endian) {
	  if (usigned) *(hbuffu+offs)=(float)valu*vol;
	  else *(hbuffs+offs)=(float)val*vol;
	}
	else {
	  *(hbuffc+offs)=(float)(val&0x00FF)*vol;
	  *(hbuffc+(++offs))=(float)((val&0xFF00)>>8)*vol;
	}
      }
      else {
	if (usigned) *(hbuffc+offs)=(guchar)((float)(valu>>8)*vol);
	else *(hbuffc+offs)=(guchar)((float)val/256.*vol);
      }
      offs++;
    }
    coffs=(gint)(coffs_f+=scale);
  }
  return frames_out;
}

//#define DEBUG_ARENDER

static void pad_with_silence(int out_fd, long oins_size, long ins_size) {
  // fill to ins_pt with zeros
  guchar *zero_buff;
  size_t sblocksize=SILENCE_BLOCK_SIZE;
  gint sbytes=ins_size-oins_size;
  int i;

#ifdef DEBUG_ARENDER
  g_print("sbytes is %d\n",sbytes);
#endif

  if (sbytes>0) {
    lseek(out_fd,oins_size,SEEK_SET);
    zero_buff=g_malloc0(SILENCE_BLOCK_SIZE);
    for (i=0;i<sbytes;i+=SILENCE_BLOCK_SIZE) {
      if (sbytes-i<SILENCE_BLOCK_SIZE) sblocksize=sbytes-i;
      dummyvar=write (out_fd,zero_buff,sblocksize);
    }
    g_free(zero_buff);
  }
  else if (sbytes<=0) {
    lseek(out_fd,oins_size+sbytes,SEEK_SET);
  }
}


void render_audio_segment(gint nfiles, gint *from_files, gint to_file, gdouble *avels, gdouble *fromtime, weed_timecode_t tc_start, weed_timecode_t tc_end, gdouble *chvol, gdouble opvol_start, gdouble opvol_end) {
  // called during multitrack rendering to create the actual audio file
  // also used for fade-in/fade-out in the clip editor (opvol_start, opvol_end)

  // in multitrack, chvol is taken from the audio mixer

  // what we will do here:
  // calculate our target samples (= period * out_rate)

  // calculate how many in_samples for each track (= period * in_rate / ABS (vel) )

  // read in the relevant number of samples for each track and convert to float

  // write this into our float buffers (1 buffer per channel per track)

  // we then send small chunks at a time to our audio volume/pan effect; this is to allow for parameter interpolation

  // the small chunks are processed and mixed, converted from float to int, and then written to the outfile


  file *outfile=mainw->files[to_file];
  size_t tbytes;
  guchar *in_buff;

  gint out_asamps=outfile->asampsize/8;
  gint out_achans=outfile->achans;
  gint out_arate=outfile->arate;
  gint out_unsigned=outfile->signed_endian&AFORM_UNSIGNED;
  gint out_bendian=outfile->signed_endian&AFORM_BIG_ENDIAN;
  short *holding_buff;
  float *float_buffer[out_achans*nfiles];

  float *chunk_float_buffer[out_achans*nfiles];

  int c,x;
  register int i;
  size_t bytes_read;
  int in_fd[nfiles],out_fd;
  gulong nframes;
  gboolean in_reverse_endian[nfiles],out_reverse_endian;
  long seekstart[nfiles];
  gchar *infilename,*outfilename;
  weed_timecode_t tc=tc_start;
  gdouble ins_pt=tc/U_SEC;
  long ins_size,cur_size;
  gdouble time=0.;
  gdouble opvol=opvol_start;
  long frames_out;

  int track;

  gint in_asamps[nfiles];
  gint in_achans[nfiles];
  gint in_arate[nfiles];
  gint in_unsigned[nfiles];
  gint in_bendian;

  gboolean is_silent[nfiles];
  gint first_nonsilent=-1;

  long tsamples=(tc_end-tc_start)/U_SEC*out_arate;

  long blocksize,zsamples,xsamples;

  void *finish_buff;

  weed_plant_t *shortcut=NULL;

  size_t max_aud_mem,bytes_to_read,aud_buffer;
  gint max_segments;
  gdouble zavel,zavel_max=0.;

  //#define DEBUG_ARENDER
  // prepare outfile stuff
  outfilename=g_strdup_printf("%s/%s/audio",prefs->tmpdir,outfile->handle);
#ifdef DEBUG_ARENDER
  g_print("writing to %s\n",outfilename);
#endif
  out_fd=open(outfilename,O_WRONLY|O_CREAT|O_SYNC,S_IRUSR|S_IWUSR);
  g_free(outfilename);
  
  cur_size=get_file_size(out_fd);

  ins_pt*=out_achans*out_arate*out_asamps;
  ins_size=((long)(ins_pt/out_achans/out_asamps+.5))*out_achans*out_asamps;

  if ((!out_bendian&&(G_BYTE_ORDER==G_BIG_ENDIAN))||(out_bendian&&(G_BYTE_ORDER==G_LITTLE_ENDIAN))) out_reverse_endian=TRUE;
  else out_reverse_endian=FALSE;
    
  // fill to ins_pt with zeros
  pad_with_silence(out_fd,cur_size,ins_size);
  sync();

#ifdef DEBUG_ARENDER
  g_print("here %d %lld %lld %d\n",nfiles,tc_start,tc_end,to_file);
#endif

  for (track=0;track<nfiles;track++) {
    // prepare infile stuff
    file *infile;

#ifdef DEBUG_ARENDER
    g_print(" track %d %d %.4f %.4f\n",track,from_files[track],fromtime[track],avels[track]);
#endif

    if (avels[track]==0.) {
      is_silent[track]=TRUE;
      continue;
    }

    is_silent[track]=FALSE;
    
    infile=mainw->files[from_files[track]];

    in_asamps[track]=infile->asampsize/8;
    in_achans[track]=infile->achans;
    in_arate[track]=infile->arate;
    in_unsigned[track]=infile->signed_endian&AFORM_UNSIGNED;
    in_bendian=infile->signed_endian&AFORM_BIG_ENDIAN;

    if ((!in_bendian&&(G_BYTE_ORDER==G_BIG_ENDIAN))||(in_bendian&&(G_BYTE_ORDER==G_LITTLE_ENDIAN))) in_reverse_endian[track]=TRUE;
    else in_reverse_endian[track]=FALSE;
    
    seekstart[track]=(long)(fromtime[track]*in_arate[track])*in_achans[track]*in_asamps[track];
    seekstart[track]=((long)(seekstart[track]/in_achans[track]/(in_asamps[track])))*in_achans[track]*in_asamps[track];
    
    zavel=avels[track]*(gdouble)in_arate[track]/(gdouble)out_arate*in_asamps[track]*in_achans[track]/sizeof(float);
    if (ABS(zavel)>zavel_max) zavel_max=ABS(zavel);

    infilename=g_strdup_printf("%s/%s/audio",prefs->tmpdir,infile->handle);
    in_fd[track]=open(infilename,O_RDONLY);
    lseek(in_fd[track],seekstart[track],SEEK_SET);
    g_free(infilename);
  }

  for (track=0;track<nfiles;track++) {
    if (!is_silent[track]) {
      first_nonsilent=track;
      break;
    }
  }


  if (first_nonsilent==-1) {
    // output silence
    long oins_size=ins_size;
    ins_pt=tc_end/U_SEC;
    ins_pt*=out_achans*out_arate*out_asamps;
    ins_size=((long)(ins_pt/out_achans/out_asamps)+.5)*out_achans*out_asamps;
    pad_with_silence(out_fd,oins_size,ins_size);
    sync();
    close (out_fd);
    return;
  }


  // we don't want to use more than MAX_AUDIO_MEM bytes
  // (numbers will be much larger than examples given

  max_aud_mem=MAX_AUDIO_MEM/(1.5+zavel_max); // allow for size of holding_buff and in_buff

  max_aud_mem=(max_aud_mem>>7)<<7; // round to a multiple of 128 

  max_aud_mem=max_aud_mem/out_achans/nfiles; // max mem per channel/track

  bytes_to_read=tsamples*(sizeof(float)); // eg. 120 (20 samples)
  max_segments=(int)((gdouble)bytes_to_read/(gdouble)max_aud_mem+1.); // max segments (rounded up) [e.g ceil(120/45)==3]
  aud_buffer=bytes_to_read/max_segments;  // estimate of buffer size (e.g. 120/3 = 40)

  zsamples=(int)(aud_buffer/sizeof(float)); // ensure whole number of samples (e.g 40 == 10 samples)

  xsamples=zsamples+(tsamples-(max_segments*zsamples)); // e.g 30 - 3 * 10 == 0

  holding_buff=(short *)g_malloc(xsamples*sizeof(short)*out_achans);
  
  for (i=0;i<out_achans*nfiles;i++) {
    float_buffer[i]=g_malloc(xsamples*sizeof(float));
  }

  finish_buff=g_malloc(RENDER_BLOCK_SIZE*out_achans*out_asamps);

  for (i=0;i<nfiles;i++) {
    for (c=0;c<out_achans;c++) {
      chunk_float_buffer[i*out_achans+c]=g_malloc(RENDER_BLOCK_SIZE*sizeof(float));
    }
  }

#ifdef DEBUG_ARENDER
  g_print("  rendering %ld samples\n",tsamples);
#endif

  while (tsamples>0) {
    tsamples-=xsamples;

    for (track=0;track<nfiles;track++) {
      if (is_silent[track]) {
	// zero float_buff
	for(c=0;c<out_achans;c++) {
	  for (x=0;x<xsamples;x++) {
	    float_buffer[track*out_achans+c][x]=0.;
	  }
	}
	continue;
      }

      // calculate tbytes for xsamples

      zavel=avels[track]*(gdouble)in_arate[track]/(gdouble)out_arate;

      tbytes=xsamples*ABS(zavel)*in_asamps[track]*in_achans[track];
      tbytes=((size_t)(tbytes/in_achans[track]/(in_asamps[track])))*in_achans[track]*(in_asamps[track]);

      in_buff=g_malloc(tbytes);

      if (zavel<0.) lseek(in_fd[track],seekstart[track]-tbytes,SEEK_SET);
    
      bytes_read=read(in_fd[track],in_buff,tbytes);

      if (zavel<0.) seekstart[track]-=tbytes;

      if (bytes_read<tbytes) memset(in_buff+bytes_read,0,tbytes-bytes_read);

      nframes=tbytes/(in_asamps[track])/in_achans[track]/ABS(zavel);

      // convert to float
      if (in_asamps[track]==1) {
	sample_move_d8_d16 (holding_buff,(guchar *)in_buff,nframes,zavel,out_achans,in_achans[track]);
      }
      else {
	sample_move_d16_d16(holding_buff,(short*)in_buff,nframes,zavel,out_achans,in_achans[track],in_reverse_endian[track]);
      }

      g_free(in_buff);

      for(c=0;c<out_achans;c++) {
	sample_move_d16_float(float_buffer[c+track*out_achans],holding_buff+c,nframes,out_achans,in_unsigned[track],chvol[track]);
      }
    }

    // now we send small chunks at a time to the audio vol/pan effect
    
    blocksize=RENDER_BLOCK_SIZE;

    for (i=0;i<xsamples;i+=RENDER_BLOCK_SIZE) {
      if (i+RENDER_BLOCK_SIZE>xsamples) blocksize=xsamples-i;

      for (track=0;track<nfiles;track++) {
	for (c=0;c<out_achans;c++) {
	  w_memcpy(chunk_float_buffer[track*out_achans+c],float_buffer[track*out_achans+c]+i,blocksize*sizeof(float));
	}
      }

      // apply audio filter(s)
      if (mainw->multitrack!=NULL) {
	// we work out the "visibility" of each track at tc
	gdouble *vis=get_track_visibility_at_tc(mainw->multitrack->event_list,nfiles,mainw->multitrack->opts.back_audio_tracks,tc,&shortcut,mainw->multitrack->opts.audio_bleedthru);
	
	// locate the master volume parameter, and multiply all values by vis[track]
	weed_apply_audio_effects(mainw->filter_map,chunk_float_buffer,mainw->multitrack->opts.back_audio_tracks,out_achans,blocksize,out_arate,tc,vis);
	
	g_free(vis);
      }
      
      // convert back to int; use out_scale of 1., since we did our resampling in sample_move_*_d16
      frames_out=sample_move_float_int((void *)finish_buff,chunk_float_buffer,blocksize,1.,out_achans,out_asamps*8,out_unsigned,out_reverse_endian,opvol);
      dummyvar=write (out_fd,finish_buff,frames_out*out_asamps*out_achans);
#ifdef DEBUG_ARENDER
      g_print(".");
#endif
      sync();

      if (mainw->multitrack==NULL) {
	time+=(gdouble)frames_out/(gdouble)out_arate;
	opvol=opvol_start+(opvol_end-opvol_start)*(time/((tc_end-tc_start)/U_SEC));
      }
      else tc+=(gdouble)blocksize/(gdouble)out_arate*U_SEC;
    }
    xsamples=zsamples;
  }

  for (i=0;i<out_achans*nfiles;i++) {
    g_free(float_buffer[i]);
  }

  g_free(finish_buff);
  g_free(holding_buff);

  // close files
  for (track=0;track<nfiles;track++) {
    if (!is_silent[track]) {
      close (in_fd[track]);
    }
    for (c=0;c<out_achans;c++) {
      g_free(chunk_float_buffer[track*out_achans+c]);
    }
  }
#ifdef DEBUG_ARENDER
  g_print("fs is %ld\n",get_file_size(out_fd));
#endif
  close (out_fd);

}


inline void aud_fade(gint fileno, gdouble startt, gdouble endt, gdouble startv, gdouble endv) {
  gdouble vel=1.,vol=1.;
  render_audio_segment(1,&fileno,fileno,&vel,&startt,startt*U_SECL,endt*U_SECL,&vol,startv,endv);
}



#ifdef ENABLE_JACK
void rec_audio_to_clip(gint fileno, gboolean is_window_grab) {
  // open audio file for writing
  file *outfile=mainw->files[fileno];
  gchar *outfilename=g_strdup_printf("%s/%s/audio",prefs->tmpdir,outfile->handle);
  gint out_bendian=outfile->signed_endian&AFORM_BIG_ENDIAN;

  mainw->aud_rec_fd=open(outfilename,O_WRONLY|O_CREAT,S_IRUSR|S_IWUSR);
  g_free(outfilename);

  mainw->jackd_read=jack_get_driver(0,FALSE);
  mainw->jackd_read->playing_file=fileno;
  mainw->jackd_read->frames_written=0;

  if ((!out_bendian&&(G_BYTE_ORDER==G_BIG_ENDIAN))||(out_bendian&&(G_BYTE_ORDER==G_LITTLE_ENDIAN))) mainw->jackd_read->reverse_endian=TRUE;
  else mainw->jackd_read->reverse_endian=FALSE;

  // start jack recording
  jack_open_device_read(mainw->jackd_read);
  jack_read_driver_activate(mainw->jackd_read);

  // in grab window mode, just return, we will call rec_audio_end on playback end
  if (is_window_grab) return;

  mainw->cancelled=CANCEL_NONE;
  mainw->cancel_type=CANCEL_SOFT;
  // show countdown/stop dialog
  d_print(_("Recording audio..."));
  do_auto_dialog(_("Recording audio"),1);
  rec_audio_end();
}



void rec_audio_end(void) {
  // recording ended

  // stop recording
  jack_close_device(mainw->jackd_read,TRUE);
  mainw->jackd_read=NULL;

  // close file
  close(mainw->aud_rec_fd);
  mainw->cancel_type=CANCEL_KILL;
}

#endif
